@@ -1,7 +1,4 @@
 from libspn_keras.losses.negative_log_joint import NegativeLogJoint
 from libspn_keras.losses.negative_log_likelihood import NegativeLogLikelihood
 
-__all__ = [
-    'NegativeLogJoint',
-    'NegativeLogLikelihood'
-]
\ No newline at end of file
+__all__ = ["NegativeLogJoint", "NegativeLogLikelihood"]
