@@ -1,9 +1,9 @@
-from .dynamic import DynamicSumProductNetwork
-from .spn import SumProductNetwork
+from .dynamic_spn import DynamicSumProductNetwork
 from .sequential_spn import SequentialSumProductNetwork
+from .spn import SumProductNetwork
 
 __all__ = [
     "DynamicSumProductNetwork",
     "SumProductNetwork",
-    "SequentialSumProductNetwork"
+    "SequentialSumProductNetwork",
 ]
